"""
Creating and visualizing summary graphs.

Author: Radoslav Chudovsky
"""
import os
import argparse
import utils


def copy_from_reference_graph(summary_graph, reference_graph):
    """Copy nodes and edges from the reference graph to the summary graph.
    Add achieved_by attribute to nodes and counter attribute to edges."""
    for node, node_data in reference_graph.nodes(data=True):
        summary_graph.add_node(node, achieved_by=set(), **node_data,
                               color=utils.GOOD_GREEN, penwidth=3)

    for from_node, to_node, edge_data in reference_graph.edges(data=True):
        if not utils.is_command_edge(edge_data):
            continue
        edge_key = edge_data['event_type'] + '_' + edge_data['tool']
        summary_graph.add_edge(from_node, to_node, **edge_data, counter=0,
                               weight=8, len=2, color=utils.GOOD_GREEN,
                               penwidth=3, key=edge_key)


def update_error_node(summary_graph, error_node_to_add, commands_to_add,
                      trainee_id):
    """
    Add an error node (if it doesn't exist) into the summary graph and
    update its in-edges with the given commands.

    :param summary_graph: summary graph to be updated
    :param error_node_to_add: (in "v_err" format) error node to be added, if it
    doesn't exist already
    :param commands_to_add: commands to be added as in-edges of the error node
    :param trainee_id: trainee's ID
    :return: None
    """
    from_node = error_node_to_add[:-4]  # remove "_err"
    error_node_to_add += "_" + trainee_id

    if not summary_graph.has_node(error_node_to_add):
        summary_graph.add_node(error_node_to_add,
                               label="By: " + trainee_id + "\\n(Error node)",
                               color=utils.BAD_RED, penwidth=2)

    for command in commands_to_add:
        utils.update_edge(summary_graph, from_node, error_node_to_add,
                          command, fontcolor=utils.BAD_RED, penwidth=2,
                          color=utils.BAD_RED)


def add_node_including_after(summary_graph, node_to_add, edge_commands,
                             trainee_id):
    """Add nodes that include "_after_" (possible nodes and error nodes
    following possible nodes) as error nodes (if they don't exist)."""
    # node_to_add looks like this -> "v_after_u" or "v_after_u_err"
    error_node = node_to_add.split('_after_')[1]
    if not error_node.endswith("_err"):
        error_node += "_err"

    update_error_node(summary_graph, error_node, edge_commands, trainee_id)


def increase_edge_counter(summary_graph, success_node, trainee_command):
    """Find in-edge of the success node that matches with the trainee's
    command and update its counter."""
    edge_key = trainee_command.event_type + '_' + trainee_command.tool
    predecessors = list(summary_graph.predecessors(success_node))

    for pred in predecessors:
        edge_data = summary_graph[pred][success_node][edge_key]

        if trainee_command.matches_edge(edge_data):
            edge_data['counter'] += 1
            return


def update_summary_graph(summary_graph, trainee_graph):
    """Update summary graph with nodes and edges from the trainee graph.
    Can be modified to exclude error nodes after the end node."""
    # uncomment to exclude error nodes after the end node
    # error_node_after_end = utils.get_end_node(summary_graph) + "_err"
    for _, to_node, edge_data in trainee_graph.edges(data=True):

        # uncomment to exclude error nodes after the end node
        # if to_node == error_node_after_end:
        #    continue

        # filter non-command edges
        if not utils.is_command_edge(edge_data):
            continue

        edge_commands = utils.create_commands_from_edge(edge_data)

        if '_after_' in to_node:
            add_node_including_after(summary_graph, to_node, edge_commands,
                                     trainee_graph.graph["name"])
            continue

        if to_node.endswith("_err"):
            update_error_node(summary_graph, to_node, edge_commands,
                              trainee_graph.graph["name"])
            continue

        # to_node is a success node
        summary_graph.nodes[to_node]["achieved_by"].add(
            trainee_graph.graph["name"])
        # edge_commands consists of one command
        increase_edge_counter(summary_graph, to_node, edge_commands[0])


def write_trainee_ids(id_set, limit=6):
    """
    Write trainees' IDs as a sting with line
    breaks so the node label is not too wide.

    :param id_set: set of trainees' IDs
    :param limit: limit of IDs on one line
    :return: IDs of trainees in a string
    """
    ids_string = '{'
    counter = 0
    id_list = list(id_set)

    for trainee_id in id_list:
        ids_string += trainee_id
        counter += 1

        if trainee_id != id_list[-1]:
            ids_string += ", "
            if counter >= limit:
                ids_string += "\\n"
                counter = 0

    return ids_string + '}'


def write_edge_labels(summary_graph):
    """Write labels of edges from the summary graph."""
    for _, to_node, edge_data in summary_graph.edges(data=True):
        if '_err' in to_node:
            edge_data['label'] = utils.write_edge_label_reduced(edge_data)
            continue
        edge_data['label'] = utils.write_edge_label_reduced(
            edge_data, True) + '({})'.format(edge_data['counter'])


def create_nx_summary_graph(reference_graph, name, trainee_graphs):
    """
    Create a summary graph (as a NetworkX graph) from trainee graphs
     located in trainee_graphs_folder.

    :param reference_graph: reference graph of the training as a NetworkX graph
    :param name: name of the summary graph that will be created
    :param trainee_graphs: list of trainee graphs (as NetworkX graphs)
    :return: created summary graph as a NetworkX graph
    """
    summary_graph = utils.nx.MultiDiGraph(name=name)

    copy_from_reference_graph(summary_graph, reference_graph)

    start_node = utils.get_start_node(summary_graph)
    for trainee_graph in trainee_graphs:
        summary_graph.nodes[start_node]["achieved_by"].add(
            trainee_graph.graph["name"])
        update_summary_graph(summary_graph, trainee_graph)

    # label success nodes with trainees' IDs who achieved them
    for node, node_data in summary_graph.nodes(data=True):
        if "_err_" not in node:
            node_label = node_data['label'] if 'label' in node_data else node
            node_data['label'] = node_label + "\n" + write_trainee_ids(node_data['achieved_by'])

    write_edge_labels(summary_graph)

    return summary_graph


def load_trainee_graphs(trainee_graphs_folder):
    """Find and load gv files with trainee graphs in graphs_folder."""
    trainee_graphs = []

    files = os.listdir(trainee_graphs_folder)

    for file in files:
        if file.endswith(".gv"):
            trainee_graph = utils.load_graph(
                os.path.join(trainee_graphs_folder, file))
            trainee_graphs.append(trainee_graph)

    return trainee_graphs


def create_and_draw_summary_graph(reference_graph_path, trainee_graphs_folder,
                                  draw_format):
    """
    Create and draw summary graph.

    :param reference_graph_path: path to the reference graph
    :param trainee_graphs_folder: folder with trainee graphs in gv files
    :param draw_format: format of the file with the drawn graph
    :return: None
    """
    if trainee_graphs_folder.endswith('/'):
        trainee_graphs_folder = trainee_graphs_folder[:-1]

    folder_name = os.path.basename(trainee_graphs_folder)

    graph_name = "summary_graph"
    if folder_name not in ('.', ""):
        graph_name += "_" + folder_name

    reference_graph = utils.load_graph(reference_graph_path)
    trainee_graphs = load_trainee_graphs(trainee_graphs_folder)
    summary_graph = create_nx_summary_graph(reference_graph, graph_name,
                                            trainee_graphs)

    result_folder = os.path.join(trainee_graphs_folder, 'summary_graph')
    if not os.path.exists(result_folder):
        os.mkdir(result_folder)

    summary_graph_filename = os.path.join(result_folder, graph_name)
    utils.draw_graph(summary_graph, summary_graph_filename, draw_format)

    print("Summary graph \"{}\" created in \"{}\"\t(from graphs in \"{}\")"
          .format(graph_name, result_folder, trainee_graphs_folder))


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(
        description="Script for creating a summary graph from trainee graphs.")
    arg_parser.add_argument("reference_graph",
                            help="path to the reference graph")
    arg_parser.add_argument("trainee_graphs_folder",
                            help="path to the folder with trainee graphs")
    arg_parser.add_argument("-f", "--format",
                            help="format of the generated graph")

    args = arg_parser.parse_args()

    create_and_draw_summary_graph(args.reference_graph,
                                  args.trainee_graphs_folder, args.format)
