"""
Creating and visualizing trainee graphs.

Author: Radoslav Chudovsky
"""
import argparse
import os
import utils


def map_edge(reference_graph, trainee_command):
    """
    Try to match trainee's command with any in-edge in the reference graph.

    If the command matches with an edge in the reference_graph,
    tuple (success node, req_nodes) will be returned, where:
    success node -- node the matched edge points to,
    req_nodes -- edge's required nodes.

    :param reference_graph: reference graph of the training
    :param trainee_command: trainee's command
    :return: tuple (success node, req_nodes)
    if trainee's command matches one of its in-edges, else None
    """
    for _, to_node, edge_data in reference_graph.edges(data=True):
        edge_commands = utils.create_commands_from_edge(edge_data)
        for edge_command in edge_commands:
            if trainee_command.matches_command(edge_command):
                if 'req_nodes' in edge_data:
                    return to_node, edge_data['req_nodes'].copy()
                return to_node, []
    return None


def add_back_transition(trainee_graph, from_node, to_node, weight):
    color = "#686868"
    trainee_graph.add_edge(from_node, to_node, key="back",
                           weight=weight, len=2, label="back\\ltransition",
                           color=color, fontcolor=color)


def add_error_node(trainee_graph, from_node, command):
    """Add an error node derived from the from_node (if it doesn't exist
    already). Update (or add) the edge with the given command leading
    from the from_node to its error node."""
    error_node = from_node + '_err'

    if not trainee_graph.has_node(error_node):
        # node_label = 'Not in the ref. graph\\nAfter: ' + from_node
        node_label = 'Not in the ref. graph\\n'
        trainee_graph.add_node(error_node, color=utils.BAD_RED, label=node_label,
                               labelfontcolor=utils.BAD_RED)

    utils.update_edge(trainee_graph, from_node, error_node, command,
                      color=utils.BAD_RED, fontcolor=utils.BAD_RED)

    # add_back_transition(trainee_graph, error_node, from_node, 2.5)


def filter_req_nodes(graph, req_nodes):
    """
    Filter req_nodes that are already in the graph.

    :param graph: graph to check
    :param req_nodes: list of nodes that will be filtered
    :return None
    """
    nodes_copy = req_nodes.copy()

    for node in nodes_copy:
        if graph.has_node(node):
            req_nodes.remove(node)


def add_edge_to_existing_node(trainee_graph, from_node, to_node, command,
                              self_edges, dashed_edges):
    """Add dashed edge leading to an existing node from the from_node, label
    the edge with the given command. Can be modified to include self-loops
    in the graph."""
    if from_node == to_node:
        if not self_edges:
            return
    else:
        if not dashed_edges:
            return

    color = utils.MAYBE_ORANGE if "_after_" in to_node else utils.GOOD_GREEN

    utils.update_edge(trainee_graph, from_node, to_node, command,
                      color=color, style="dashed", fontcolor=color, weight=1)


def add_success_node(trainee_graph, curr_success_node, node_to_add, command, label):
    """Add a new success node along with an edge leading from the current
    success node to the new one."""
    trainee_graph.add_node(node_to_add, color=utils.GOOD_GREEN, penwidth=3, label=label)

    utils.update_edge(trainee_graph, curr_success_node, node_to_add, command,
                      color=utils.GOOD_GREEN, penwidth=2.5, weight=10, len=2)


def add_possible_node(trainee_graph, from_node, node_to_add, command,
                      missing_nodes, reference_graph):
    """Add a new possible node to the graph labeled with missing nodes."""
    tried_to_reach = node_to_add.split("_after_")[0]
    if 'label' in reference_graph.nodes()[tried_to_reach]:
        tried_to_reach = reference_graph.nodes()[tried_to_reach]['label']
    node_label = "Tried to reach: {}\\nMissing nodes: {}" \
        .format(tried_to_reach, missing_nodes)
    trainee_graph.add_node(node_to_add, color=utils.MAYBE_ORANGE, penwidth=2.5,
                           label=node_label)

    utils.update_edge(trainee_graph, from_node, node_to_add, command,
                      color=utils.MAYBE_ORANGE, penwidth=2, weight=5, len=2)


def write_edge_labels(trainee_graph):
    """Write labels of trainee graph's command edges."""
    for _, _, edge_data in trainee_graph.edges(data=True):
        if utils.is_command_edge(edge_data):
            edge_data['label'] = utils.write_edge_label_extended(edge_data)


def write_graph_label(trainee_graph, reference_graph):
    percentage = 100*trainee_graph.graph['sn_counter']/(len(reference_graph.nodes()) - 1)
    label = "Trainee: {}\\lReached {} of {} available success nodes ({}%)\\l"\
        .format(trainee_graph.graph['name'], trainee_graph.graph['sn_counter'],
                len(reference_graph.nodes()) - 1, round(percentage, 2))

    if 'graph' in reference_graph.graph and 'training' in reference_graph.graph['graph']:
        label += "Training: {}\\l".format(reference_graph.graph['graph']['training'])

    colors = "Green nodes: Successfully reached states from the reference graph\\l" \
             "Red nodes: Commands not present in the reference graph\\l" \
             "Orange nodes: Attempts to reach a green node but some required states were missing\\l" \
             "Gray nodes: Reference solution to complete the rest of the training\\l"

    label += "\\n" + colors

    trainee_graph.graph['graph'] = {'label': label, 'labelloc': 'top', 'labeljust': 'left',
                                    'fontsize': 25}


def add_unreached_nodes(trainee_graph, reference_graph, last_reached):
    last_node = last_reached
    color = "#686868"

    for _, to_node, edge_data in reference_graph.edges(data=True):
        if not trainee_graph.has_node(to_node):
            node_label = to_node
            if 'label' in reference_graph.nodes()[to_node]:
                node_label = reference_graph.nodes()[to_node]['label']
            trainee_graph.add_node(to_node, label="(unreached)\\n" + node_label, color=color,
                                   fontcolor=color)

            edge_key = trainee_graph.add_edge(last_node, to_node, **edge_data, color=color,
                                              label=utils.write_edge_label_extended(edge_data),
                                              fontcolor=color)
            # add required nodes to the edge label
            req_nodes_label = 'Req. nodes: [' + utils.write_req_nodes(edge_data['req_nodes'],
                                                                      reference_graph) + ']\\l'
            trainee_graph[last_node][to_node][edge_key]['label'] += req_nodes_label
            last_node = to_node


def highlight_final_node(trainee_graph, final_node):
    trainee_graph.nodes[final_node]['label'] = "(last node)\\n" + trainee_graph.nodes[final_node]['label']
    trainee_graph.nodes[final_node]['style'] = 'filled'

    color = trainee_graph.nodes[final_node]['color']
    if color == utils.GOOD_GREEN:
        trainee_graph.nodes[final_node]['fillcolor'] = '#00CC00'
    elif color == utils.BAD_RED:
        trainee_graph.nodes[final_node]['fillcolor'] = '#FF8080'
    else:
        trainee_graph.nodes[final_node]['fillcolor'] = '#FFA333'


def create_nx_trainee_graph(reference_graph, commands, trainee_id,
                            self_edges=True, dashed_edges=True):
    """
    Create a trainee graph (as a NetworkX graph) from trainee's
    commands and a reference_graph.

    Can be modified to stop modeling after reaching the end node
    or to exclude dashed edges.

    :param reference_graph: reference graph of the training (as a
    NetworkX graph_
    :param commands: list of trainee's commands (class Command)
    :param trainee_id: trainee's ID
    :param self_edges: enable/disable self-edges
    :param dashed_edges: enable/disable dashed edges
    :return: created trainee graph as NetworkX graph
    """
    trainee_graph = utils.nx.MultiDiGraph(name=trainee_id, sn_counter=0)

    start_node = utils.get_start_node(reference_graph)
    trainee_graph.add_node(start_node, color=utils.GOOD_GREEN, penwidth=3)

    # uncomment to stop modeling after reaching the end node
    # end_node = utils.get_end_node(reference_graph)

    current_success_node = start_node
    current_possible_node = None
    # last reached node
    final_node = start_node

    if self_edges is None:
        self_edges = True

    for command in commands:
        mapping_result = map_edge(reference_graph, command)
        # last reached success or possible node
        last_node = current_success_node if current_possible_node is None \
            else current_possible_node

        # adding an error node
        if mapping_result is None:
            add_error_node(trainee_graph, last_node, command)
            final_node = last_node + '_err'
            continue

        mapped_node, req_nodes = mapping_result
        filter_req_nodes(trainee_graph, req_nodes)

        # mapped_node is already in the graph
        if trainee_graph.has_node(mapped_node):
            # comment next 2 commands to skip adding dashed edges to success nodes
            add_edge_to_existing_node(trainee_graph, last_node, mapped_node,
                                      command, self_edges, dashed_edges)
            final_node = last_node
            continue

        # adding a success node -- all req_nodes are in the graph
        if len(req_nodes) == 0:
            label = mapped_node
            if 'label' in reference_graph.nodes()[mapped_node]:
                label = reference_graph.nodes()[mapped_node]['label']
            add_success_node(trainee_graph, current_success_node, mapped_node,
                             command, label)
            if current_possible_node is not None:
                # add_back_transition(trainee_graph, current_possible_node,
                #                    current_success_node, 4)
                current_possible_node = None
            current_success_node = mapped_node
            final_node = mapped_node
            trainee_graph.graph['sn_counter'] += 1
            # uncomment to stop modeling after reaching the end node
            # if current_success_node == end_node:
            #     break
            continue

        # adding a possible node -- some req_nodes are missing
        mapped_node += "_after_" + current_success_node
        if trainee_graph.has_node(mapped_node):
            # comment next line to skip adding dashed edges to possible nodes
            add_edge_to_existing_node(trainee_graph, last_node, mapped_node,
                                      command, self_edges, dashed_edges)
            continue
        req_nodes_string = utils.write_req_nodes(req_nodes, reference_graph)
        add_possible_node(trainee_graph, last_node, mapped_node, command,
                          req_nodes_string, reference_graph)
        current_possible_node = mapped_node
        final_node = mapped_node

    # highlight the final node
    highlight_final_node(trainee_graph, final_node)

    for _, _, edge_data in trainee_graph.edges(data=True):
        if utils.is_command_edge(edge_data):
            edge_data['label'] = utils.write_edge_label_extended(edge_data)

    write_graph_label(trainee_graph, reference_graph)

    add_unreached_nodes(trainee_graph, reference_graph, current_success_node)

    return trainee_graph


def create_and_draw_trainee_graph_json(reference_graph, command_log_path,
                                       result_folder, draw_format, self_edges,
                                       dashed_edges):
    """
    Create and draw a trainee graph from the given reference graph
    and json log file and save it to the specified result_folder.

    :param reference_graph: reference graph as a NetworkX graph
    :param command_log_path: path to a command log file
    :param result_folder: folder  where to save the generated graph
    :param draw_format: file format of the drawn file
    :param self_edges: enabling/disabling self-edges
    :param dashed_edges enabling/disabling dashed edges
    :return created trainee graph
    """
    command_log_base = os.path.basename(command_log_path)
    command_log_name = os.path.splitext(command_log_base)[0]

    commands = utils.log_parser.parse_commands_json(command_log_path)

    trainee_graph = create_nx_trainee_graph(reference_graph, commands,
                                            command_log_name, self_edges,
                                            dashed_edges)

    trainee_graph_filename = command_log_name + '_trainee_graph'
    trainee_graph_location = os.path.join(result_folder,
                                          trainee_graph_filename)
    utils.draw_graph(trainee_graph, trainee_graph_location, draw_format)
    print("Trainee graph \"{}\" created in \"{}\"\t(from log: \"{}\")"
          .format(trainee_graph_filename, result_folder, command_log_base))
    return trainee_graph


def create_and_draw_trainee_graphs_csv(reference_graph, command_log_path,
                                       result_folder, draw_format, self_edges,
                                       dashed_edges):
    """
    Create and draw a trainee graphs from the given reference graph
    and csv log file and save it to the specified result_folder.

    :param reference_graph: reference graph as a NetworkX graph
    :param command_log_path: path to a command log file
    :param result_folder: folder  where to save the generated graphs
    :param draw_format: file format of the drawn files
    :param self_edges: enabling/disabling self-edges
    :param dashed_edges enabling/disabling dashed edges
    :return created trainee graphs from the given csv file
    """
    command_log_base = os.path.basename(command_log_path)
    command_log_name = os.path.splitext(command_log_base)[0]

    result_folder = os.path.join(result_folder, command_log_name)
    if not os.path.exists(result_folder):
        os.mkdir(result_folder)

    commands = utils.log_parser.parse_commands_csv(command_log_path)

    trainee_graphs = []
    for student_id in commands:
        trainee_graph = create_nx_trainee_graph(reference_graph, commands[student_id],
                                                student_id, self_edges, dashed_edges)
        trainee_graphs.append(trainee_graph)
        trainee_graph_filename = student_id + '_trainee_graph'
        trainee_graph_location = os.path.join(result_folder, trainee_graph_filename)
        utils.draw_graph(trainee_graph, trainee_graph_location, draw_format)
        print("Trainee graph \"{}\" created in \"{}\"\t(from log: \"{}\")"
              .format(trainee_graph_filename, result_folder, command_log_base))
    return trainee_graphs


def compute_graphs(trainee_graphs):
    graphs_n = 0
    edges_n = 0
    nodes_n = 0

    for graph in trainee_graphs:
        graphs_n += 1
        edges_n += len(graph.edges())
        nodes_n += len(graph.nodes())
        
    if graphs_n == 0:
    	avg_nodes = 0
    	avg_edges = 0
    else:
    	avg_nodes = round(nodes_n/graphs_n, 2)
    	avg_edges = round(edges_n/graphs_n, 2)

    print("Graphs: {}".format(graphs_n))
    print("Nodes: {} (Avg: {})".format(nodes_n, avg_nodes))
    print("Edges: {} (Avg: {})".format(edges_n, avg_edges))


def create_trainee_graphs(reference_graph_path, logs_folder, draw_format, self_edges, dashed_edges):
    """
    Create and draw trainee graphs from the given trainee graph
    and command logs. The reference graph is loaded from the
    reference_graph_path. Logs in json or jsonl file formats are
    loaded from the logs_directory.

    The trainee graphs will be saved in a folder called "trainee_graphs"
    located in the folders with logs. If such a folder doesn't exist,
    it will be created.

    :param reference_graph_path: path to a reference graph
    :param logs_folder: folder where command logs of trainees are located
    :param draw_format: format of the file with the drawn graph
    :param self_edges enabling/disabling self-edges
    :param dashed_edges enabling/disabling dashed edges
    :return: None
    """
    reference_graph = utils.load_graph(reference_graph_path)

    results_folder = os.path.join(logs_folder, "trainee_graphs")
    if not os.path.exists(results_folder):
        os.mkdir(results_folder)

    files = os.listdir(logs_folder)
    trainee_graphs = []
    for file in files:
        if file.lower().endswith('.json') or file.lower().endswith('.jsonl'):
            log_file_path = os.path.join(logs_folder, file)
            graph = create_and_draw_trainee_graph_json(reference_graph, log_file_path, results_folder,
                                                       draw_format, self_edges, dashed_edges)
            trainee_graphs.append(graph)
        if file.lower().endswith('.csv'):
            log_file_path = os.path.join(logs_folder, file)
            graphs = create_and_draw_trainee_graphs_csv(reference_graph, log_file_path, results_folder,
                                                        draw_format, self_edges, dashed_edges)
            trainee_graphs += graphs

    compute_graphs(trainee_graphs)


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(
        description="Script for creating trainee graphs from command logs.")

    arg_parser.add_argument("reference_graph",
                            help="path to the reference graph")
    arg_parser.add_argument("command_logs_folder",
                            help="path to the folder with command logs")
    arg_parser.add_argument("-f", "--format",
                            help="format of generated graphs")

    arg_parser.add_argument("--disable-self-edges", action='store_false',
                            help="disable edges beginning and ending in the same node")

    arg_parser.add_argument("--disable-dashed-edges", action='store_false',
                            help="disable dashed edges leading to already existing nodes")

    args = arg_parser.parse_args()
    create_trainee_graphs(args.reference_graph, args.command_logs_folder,
                          args.format, args.disable_self_edges, args.disable_dashed_edges)
