"""
Parsing of command logs and the definition of Command class.

Author: Radoslav Chudovsky, 469069
"""
import json
import csv


class Command:
    """
    Class representing an executed command.

    Attributes:
        event_type = type of the command
        tool = name of the used tool
        options = options used with the tool
    """
    def __init__(self, event_type, tool, options):
        self.event_type = event_type
        self.tool = tool
        self.options = options

    def matches_command(self, other_command):
        """Check if two commands match -- their event_type and tool are
        the same, and the options of both commands are equal sets."""
        if self.event_type != other_command.event_type or \
                self.tool != other_command.tool:
            return False
        return self.options_match(other_command.options)

    def matches_edge(self, edge_data):
        """Check if command matches the edge -- event_type and tool are
        the same, and set(options) is in the options sets of the edge."""
        if self.event_type != edge_data['event_type'] or \
                self.tool != edge_data['tool']:
            return False
        for other_options in edge_data['options']:
            if self.options_match(other_options):
                return True
        return False

    def options_match(self, other_options):
        if len(self.options) != len(other_options):
            return False
        options_index = 0
        for other_option in other_options:
            if other_option == '*':
                options_index += 1
                continue
            if other_option != self.options[options_index]:
                return False
            options_index += 1
        return True


def create_command_json(command_dict):
    """
    Create an instance of Command class from data parsed by json.loads().

    :param command_dict: parsed line in dictionary by json.loads()
    :return command (class Command) created from the given data
    :return command (class Command) created from the given data
    """
    event_type = command_dict['programname']

    # if event_type == 'msf4-history':
    #     command_str = command_dict['message']
    # else:
    #     command_match = re.search('cmd="(.*?)"', command_dict['message'])
    #     command_str = command_match.group(1)
    command_str = command_dict['cmd']

    command_split = command_str.split()
    tool = command_split[0]
    options = command_split[1::]

    # if a string contains a word that starts and ends
    # with single quotes (\'), split() puts the whole word
    # into double quotes (\") and creates lists like this:
    #
    # "one 'two'".split() -> ['one', "'two'"]
    #
    # therefore, the following code swaps \' for \"
    for i in range(0, len(options)):
        options[i] = options[i].replace('\'', '\"')

    command = Command(event_type, tool, options)

    return command


def parse_commands_json(logs_filename):
    """
    Parse commands from the given log file into Command class instances.

    :param logs_filename: file with the logs of commands
    :return: list of Command class instances
    """
    commands = []

    with open(logs_filename) as file:
        for line in file:
            if not line.startswith('{'):
                line = line[43:]  # remove prefix
            line_dict = json.loads(line)
            command = create_command_json(line_dict)
            commands.append(command)

    return commands


def create_command_csv(row):
    event_type = 'command'

    input_split = row[4].split()
    tool = input_split[0]
    options = input_split[1::]

    command = Command(event_type, tool, options)
    return command


def parse_commands_csv(logs_filename):
    commands = {}

    with open(logs_filename) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        for row in csv_reader:
            student_id = (row[5] if len(row[6]) == 0 else row[6])\
                .replace('@file-wrangler', '')
            if student_id not in commands:
                commands[student_id] = []
            commands[student_id].append(create_command_csv(row))

    return commands
