"""
Utility methods for creating both the trainee graphs and the summary graphs.

Author: Radoslav Chudovsky, 469069
"""
from ast import literal_eval
import networkx as nx
import log_parser


# colors of nodes and edges
GOOD_GREEN = '#008000'  # success nodes
MAYBE_ORANGE = '#FF8C00'  # possible nodes
BAD_RED = '#FF0000'  # error nodes


def is_command_edge(edge_data):
    """Check if the edge is a command edge -- contains a command."""
    return 'event_type' in edge_data and 'tool' in edge_data \
           and 'options' in edge_data


def create_commands_from_edge(edge_data):
    """
    Return commands (class Command) represented by the edge.
    Each options item in the options list is a separate class Command entity.

    :param edge_data: edge attributes
    :return: list of commands (class Command) represented by the edge
    """
    edge_commands = []

    event_type = edge_data['event_type']
    tool = edge_data['tool']

    for options in edge_data['options']:
        command = log_parser.Command(event_type, tool, options)
        edge_commands.append(command)

    return edge_commands


def update_edge(graph, from_node, to_node, command, **edge_atr):
    """
    Create an edge between from_node and to_node with the given command
    in the given graph. If the edge with the given command already exists,
    update its options attribute.

    Options attribute in the edges is an ordered list of the first occurrences
    of the options that were used with the given command.

    Each edge between two nodes needs to have a unique key
    (event_type + tool in our case).

    :param graph: graph to be updated
    :param from_node: from-node of the edge
    :param to_node: to-node of the edge
    :param command: command (class Command) to be represented by the edge
    :param edge_atr: additional edge attributes
    :return: None
    """
    key = command.event_type + '_' + command.tool

    if graph.has_edge(from_node, to_node, key):
        edge_command_options = graph[from_node][to_node][key]['options']
        if command.options not in edge_command_options:
            edge_command_options.append(command.options)
    else:
        graph.add_edge(from_node, to_node, key, event_type=command.event_type,
                       tool=command.tool, options=[command.options],
                       **edge_atr)


def get_start_node(graph):
    """
    Return a start node of the given graph.

    Potential start node will be searched in this order:
    a node called "start", if it doesn't exist ->
    the first node with no in-edges, if it doesn't exist ->
    the first node defined in the textual graph specification.

    :param graph: given graph
    :return: start node of the given graph
    """
    if graph.has_node("start"):
        return "start"

    for node in graph.nodes():
        if len(graph.in_edges(node)) == 0:
            return node

    nodes = list(graph.nodes())
    if len(nodes) == 0:
        return None
    return nodes[0]


def get_end_node(graph):
    """
    Return end node of the given graph.

    Potential end node will be searched in this order:
    a node called "end", if it doesn't exist ->
    the first node with no out-edges, if it doesn't exist ->
    the last node defined in the textual graph specification.

    :param graph: given graph
    :return: end node
    """
    if graph.has_node("end"):
        return "end"

    for node in graph.nodes():
        if len(graph.out_edges(node)) == 0:
            return node

    nodes = list(graph.nodes())
    if len(nodes) == 0:
        return None

    return nodes[-1]


def write_options(all_options_list, length_limit=30):
    """
    Write command options into a string with line breaks, so the
    edge label isn't too wide.

    :param all_options_list: python list of options
    :param length_limit: approximated line length
    :return: commands options in a string
    """
    all_options_string = ""
    current_length = 0

    for options in all_options_list:
        all_options_string += '['
        # for option in options:
        for i in range(0, len(options)):
            option = options[i]
            option = option.replace("\\", "\\\\")  # if someone writes '\' as an argument
            all_options_string += "\'" + option + "\'"
            current_length += len(option)
            if i != len(options) - 1:
                if current_length >= length_limit:
                    current_length = 0
                    all_options_string += ',\\l'
                else:
                    all_options_string += ', '
        all_options_string += ']'

        if options != all_options_list[-1]:
            if current_length >= length_limit:
                current_length = 0
                all_options_string += ';\\l'
            else:
                all_options_string += '; '

    if all_options_string.endswith("\\l"):
        return all_options_string
    return all_options_string + "\\l"


def write_edge_label_extended(edge_data):
    """Write an extended edge label including all information."""
    return "Type: " + edge_data['event_type'] \
           + "\\lTool: " + edge_data['tool']  \
           + "\\lOpts: " + write_options(edge_data['options'])


def write_edge_label_reduced(edge_data, include_options=False):
    """Write a reduced edge label including the name of the used tool,
    optionally include commands."""
    if include_options:
        return edge_data['tool'] + ' ' + write_options(edge_data['options'], 1)
    return edge_data['tool'] + '\\l'


def write_req_nodes(req_nodes, reference_graph, length_limit=18):
    req_nodes_string = ""
    current_length = 0

    for req_node in req_nodes:
        if 'label' in reference_graph.nodes()[req_node]:
            label = reference_graph.nodes()[req_node]['label']
            req_nodes_string += label
            current_length += len(label)
        else:
            req_nodes_string += req_node
            current_length += len(req_node)

        if req_node != req_nodes[-1]:
            if current_length >= length_limit:
                current_length = 0
                req_nodes_string += ',\\n'
            else:
                req_nodes_string += ', '

    return req_nodes_string


def load_graph(filename):
    """
    Load a graph from the given file. Try to evaluate lists from
    strings to Python lists (if it has any).

    One-word attributes and their values can be written in a
    DOT graph with quotes or without them. When reading the
    file, Pydot adds quotes to every attribute name and value.

    e.g. -> Pydot would read "command" (in a DOT file) as
    ""command""

    Therefore, we need to check if the values don't have extra
    quotes and remove them.

    :param filename: filename of graph in a DOT format
    :return: given graph converted to a NetworkX graph
    """
    graph = nx.nx_pydot.read_dot(filename)

    for _, _, edge_data in graph.edges(data=True):
        # edge_data is a dictionary
        # key -> attribute name
        # key value -> attribute value
        for key in edge_data:

            # additional quotes in the key
            if key.startswith('"') and key.endswith('"'):
                new_key = key[1:-1]
                edge_data[new_key] = edge_data.pop(key)
                key = new_key

            # Pydot adds additional '\'
            edge_data[key] = edge_data[key].replace('\\', '')
            value = edge_data[key]
            # additional quotes in the key value
            if value.startswith('"') and value.endswith('"'):
                edge_data[key] = value[1:-1]

            # evaluate lists from string to python lists
            if key == 'options':
                edge_data['options'] = literal_eval(edge_data['options'])
                assert isinstance(edge_data['options'], list)

            if key == 'req_nodes':
                edge_data['req_nodes'] = literal_eval(edge_data['req_nodes'])
                assert isinstance(edge_data['req_nodes'], list)

    return graph


def draw_graph(graph, name, draw_format='pdf', save_as_gv=True):
    """
    Convert NetworkX graph into a DOT graph and draw it in the
    specified format (PDF by default).

    :param graph: NetworkX graph that will be visualized
    :param name: name of the created file(s)
    :param draw_format: desired format of the drawn graph, PDF by default
    :param save_as_gv: save in a textual format too, True by default
    :return: None
    """
    if draw_format is None:
        draw_format = 'pdf'

    pydot_graph = nx.nx_pydot.to_pydot(graph)

    if save_as_gv:
        pydot_graph.write(name + ".gv")

    pydot_graph.write(name + '.' + draw_format, format=draw_format)
